
const EventEmitter = require("eventemitter3");
const uuid = require("node-uuid");

class Node {
  constructor(options, eventemitter) {
    this.id = uuid.v1();
    this.options = options;
    this.ee = eventemitter ? eventemitter : new EventEmitter();
    this.observers = [];
  }

  push(signal) {
    setImmediate(() => {
      this.onReceive(signal);
    });
    return this;
  }

  onReceive(signal) {
    this.invokeObservers("onReceive", signal);

    if (this.isErrorSignal(signal)) {
      this.onError(signal);
      this.request();
    } else if (this.isEndSignal(signal)) {
      try {
        this.onEnd(signal);
      } catch (error) {
        this.throwError(error, signal);
      }
    } else {
      try {
        this.onSignal(signal);
      } catch (error) {
        this.throwError(error, signal);
      }
    }
    return this;
  }

  onSignal(signal) {
    this.send(signal);
    return this;
  }

  onError(error) {
    this.send(error);
    return this;
  }

  onEnd(signal) {
    this.send(signal);
    return this;
  }

  send(signal) {
    this.invokeObservers("send", signal);
    setImmediate(() => {
      this.ee.emit("outgoing-" + this.id, signal);
    });
    return this;
  }

  observe(observer) {
    this.observers.push(observer);
    return this;
  }

  to(downstream) {
    this.invokeObservers("to", downstream);
    this.ee.on("outgoing-" + this.id, signal => {
      downstream.push(signal);
    });
    // for pull model
    downstream.ee.on("request-" + downstream.id, cmd => {
      this.pull(cmd);
    });
    downstream.from(this);
    return downstream;
  }

  pull(cmd) {
    setImmediate(() => {
      this.onRequest(cmd);
    });
    return this;
  }

  onRequest(cmd) {
    this.invokeObservers("onRequest", cmd);
    this.request(cmd);
    return this;
  }

  request(cmd) {
    this.invokeObservers("request", cmd);
    setImmediate(() => {
      this.ee.emit("request-" + this.id, cmd);
    });
    return this;
  }

  observeCommand(observer) {
    this.observers.push(observer);
    return this;
  }

  from(upstream) {
    this.invokeObservers("from", upstream);
    return this;
  }

  isErrorSignal(signal) {
    return signal instanceof Error;
  }

  isEndSignal(signal) {
    //$FlowIgnore
    return signal === Node.END;
  }

  throwError(error, signal) {
    //$FlowIgnore
    error.signal = signal;
    return this.send(error);
  }

  invokeObservers(when, ...data) {
    this.observers.forEach(fn => {
      fn(this, when, data);
    });
  }
}

//$FlowIgnore
Node.END = "__SIGNAL_END__";

module.exports = Node;